# FILE: buildlib.tcl
# A library of Tcl procs which supports the script ../build.tcl .
#
######################################################################
# CONFIGURATION SUPPORT
######################################################################
# proc Default var val
#  If var already exists, do nothing.  If not, set the value of var to val.
#  Returns whatever value var ends up with.
proc Default {varref val} {
  upvar $varref var
  if {![info exists var]} {
    set var $val
  }
  return $var
}

# proc EvaluateScript script
#	A more elaborate "source" which includes calling information in
#	its error messages.
proc EvaluateScript {script} {
  if [catch {uplevel #0 source $script} msg] {
    return -code error "[info script]: Unable to evaluate $script .\n$msg"
  }
}

# proc AllDefined list
#	Returns true iff all variables in the list are set.
proc AllDefined {list} {
    set result 1
    foreach varref $list {
        upvar $varref var
        set result [expr $result && [info exists var]]
    }
    return $result
}

# proc CommaList list
#	Returns a string containing the elements of list separated by
#	commas and a final "and".
proc CommaList {list} {
    set last [expr [llength $list] - 1]
    set final [lindex $list $last]
    lreplace $list $last $last "and $final"
    return [join $list ", "]
}

# proc ReportMissingValues varlist
#	Print a report to stderr that varlist includes an undefined
#	variable, and die.
proc ReportMissingValues {varlist} {
    set caller [info script]
    puts stderr "$caller: At least one of the variables"
    puts stderr "$caller: [CommaList $varlist]"
    puts stderr "$caller: could not be supplied with a default value."
    puts stderr "$caller: Please see ./build/config.tcl for"
    puts stderr "$caller: configuration instructions."
    flush stderr
    exit 1
}

proc AppendUnique {listref value} {
  upvar $listref list
  if {[lsearch -exact $list $value] < 0} {
    lappend list $value
  }
}

proc AppendUniqueExistingReadableDirectory {listref dirref} {
  upvar $listref list
  upvar $dirref dir
  if [info exists dir] {
    if {[file isdirectory $dir] && [file readable $dir]} {
      AppendUnique list $dir
    }
  }
}

# Something changed in the way the Tcl command 'exec' passes the grouping 
# of arguments within Tcl to become the grouping of arguments within the 
# shell.  This matters when the arguments contain spaces, which is often 
# the case when arguments are pathnames on a Windows platform.
#
# Although I haven't nailed down the precise difference, an example
# Tcl script which behaves differently in Tcl 7.5 than in Tcl 7.6 is
#
#    exec {C:\WINDOWS\COMMAND.COM} /c copy README "\\Program Files"
#
# which works in Tcl 7.6 (on Windows 95), but fails in Tcl 7.5 with an
# error message 'Too many parameters', indicating that the grouping
# provided by the double quotes is not being passed to the shell.
#
# The grouping can be forced in Tcl 7.5 by entering the low circles of
# Quoting Hell to add another layer of quotes to be passed to the shell
#
#    exec {C:\WINDOWS\COMMAND.COM} /c copy README "\"\\Program Files\""
#
# which works in Tcl 7.5, but that attempt fails in Tcl 7.6 with an error
# message 'Too many parameters'.  Possibly the explicitly added quotes are
# canceling some added by exec in Tcl 7.6 and later? 
#
# The following proc, BuildExec gathers all calls to the Tcl command 'exec' 
# to one place so as the solution is refined it will be applied throughout 
# the build scripts.  BuildExec takes one argument -- a list.  Each element
# of the list will be passed as a separate argument to a call to the Tcl
# command 'exec' -- which will then in turn become a separate word in the
# command line to be executed.  See the exec(n) man page for details.
# Quoting appropriate the the Tcl version and/or platform will be added so that
# BuildExec is more nearly platform- and Tcl-version-independent than the
# Tcl command 'exec'.
proc BuildExec {cmdlist} {
  global tcl_version
  if {[llength $cmdlist]} {
    set quotedcmdlist {}
    foreach word $cmdlist {
      # If a word contains whitespace, it needs quoting at the shell level.
      # On Windows, Tcl 7.6 and later provides the necessary quoting.
      # For Tcl 7.5, we need to add the quoting explicitly
      # This code also adds quoting on Unix platforms, but that should be
      # harmless.  If not, add a platform check here.
      if {[regexp "\[ \t\r\n\]" $word] && ($tcl_version < 7.6)} {
        lappend quotedcmdlist \"$word\"
      } else {
        lappend quotedcmdlist $word
      }
    }
    # Echo the command we're about to execute...
    puts [join $quotedcmdlist]
    # ...and pass it to the shell for execution
    return [eval exec $quotedcmdlist]
  }
  return
}

# Several subcommands were added to the Tcl command 'file' in the 7.6 release
# which are useful for manipulating files: delete, copy, rename, mkdir.  In 
# order to have these abilities in Tcl 7.5, we need to evaluate shell commands.
# This proc hides away all the platform details of forming the proper list of
# arguments to BuildExec for the shell in use.
proc ShellPrompt {cmd args} {
  global env tcl_platform
  if {[string match windows $tcl_platform(platform)]} {
    if {[info exists env(COMSPEC)]} {
      set prefix [list $env(COMSPEC) /c]
    } elseif {[info exists env(ComSpec)]} {
      set prefix [list $env(ComSpec) /c]
    } elseif {[string match "windows nt" $OS]} {
      set prefix [list cmd.exe /c]
    } else {
      set prefix [list command.com /c]
    }
    set formattedargs {}
    foreach file $args {
      regsub -all / $file {\\} newfile
      lappend formattedargs $newfile
    }
  } elseif {[string match unix $tcl_platform(platform)]} {
    set prefix {}
    set formattedargs $args
  } else {
    error "Shell prompt not available on platform $tcl_platform(platform)"
  }
  if {[catch {BuildExec [concat $prefix $cmd $formattedargs]} msg]} {
    error "Error: $msg
evaluating shell command '[join [concat $cmd $formattedargs]]'"
  }
}

# On Windows NT platforms, the DOS command 'del' may take multiple arguments,
# like its Unix counterpart 'rm'.  On Windows 95, though, 'del' takes only one 
# argument, and files must be deleted one at a time (modulo wildcards).  This
# proc bridges the gap.
proc DOSRM {args} {
  foreach file $args {
    if {[file exists $file]} {
      ShellPrompt del $file
    }
  }
}

# On Windows platforms, the DOS command 'copy' takes only two arguments,
# source and destination, and files must be copied one at a time (modulo
# wildcards).  The Tcl 7.6+ counterpart, 'file copy', accepts that syntax,
# but also accepts multiple arguments, interpreting the last as a destination
# directory, and all others as files to copy to that destination directory.
# This proc bridges the gap.
proc DOSCP {args} {
  if {[llength $args] < 2} {
    error "DOSCP needs source and destination arguments"
  }
  set destdir [lindex $args [expr [llength $args] - 1]]
  if {[file isdirectory $destdir]} {
    foreach file [lrange $args 0 [expr [llength $args] - 2]] {
      if {[file exists $file]} {
        ShellPrompt copy $file $destdir
      }
    }
  } else {
    if {[llength $args] == 2} {
      set source [lindex $args 0]
      ShellPrompt copy $source $destdir
    } else {
      error "Last of multiple arguments to DOSCP must be directory"
    }
  }
}

# Proc to extract a recursive list of all #include ".*" files
# Results are stored in global var ArrInclDep.
# NOTE 1: Does not search into '#include <.*>' (system) files.
# NOTE 2: Aborts with an error message if a cyclic dependency is detected.
proc FindIncludeDep { filename } {
    global ArrInclDep INCLUDEDIRS PLATFORM
    set filename [string trim $filename]
    if [info exists ArrInclDep($filename)] { return $ArrInclDep($filename) }
    if [info exists ArrInclDep($filename,WoRkInG)] {
	puts stderr "FATAL ERROR: Cyclic #include dependency detected \
		on file $filename"
	exit 1
    }
    set worklist {}     ;# Dep lists may be empty!
    set ArrInclDep($filename,WoRkInG) {} ;# Guard against cyclic dependencies
    # Ought to use file join
    if {[string compare [file dirname $filename] .]} {
      set pathname $filename
    } else {
      if {[string compare $PLATFORM windows]==0} {
        set pathname [file join [file dirname \
                           [info script]] [file tail $filename]]
      } else {
        set pathname [file dirname [info script]]/[file tail $filename]
      }               
    }
    if [catch {open $pathname "r"} fileid] {
      puts stderr "ERROR: Can't open #include dependency \"$pathname\""
      flush stderr   ;# Safety
      return [set ArrInclDep($filename) {}]
    }
    while { [gets $fileid line] >= 0 } {
	if [regexp "^\[ \t\]*#\[ \t\]*include\[ \t\]+\"(\[^\"\]+)\"" \
		$line match newfile] {
	    set newfile [string trim $newfile]
            # Search for newfile in the same directory as filename,
            # and also in INCLUDEDIRS
            set newpath [file dirname $filename]/$newfile
            if {![file readable $newpath]} {
              foreach dir $INCLUDEDIRS {
                # Ought to use file join
                if {[string compare $PLATFORM windows]==0} {
                  set newpath [file join $dir $newfile]
                } else {
                  set newpath $dir/$newfile
                }               
                if [file readable $newpath] {
                  break
                }
              }
              if {![file readable $newpath]} {
                puts stderr "ERROR: Can't find #include dependency \"$newfile\""
                puts stderr "located in file $filename ."
                puts stderr "Searched in [file dirname $filename], as well as"
                puts stderr "$INCLUDEDIRS ."
                flush stderr   ;# Safety
                return [set ArrInclDep($filename) {}]
              }
            }
	    # See if $newpath is already in the list.  (We don't want
	    # infinite recursion!)
	    if {[string compare $filename $newpath]==0} continue
	    if {[lsearch -exact $worklist $newpath]<0} {
		# $newpath is a new file
		lappend worklist $newpath
		set worklist [concat \
			$worklist [FindIncludeDep $newpath]]
		# Note: ArrInclDep($filename) must not be defined (set)
		# until *after* these recursive calls.
	    }
	} elseif [regexp -nocase \
		"/\\\*\[ \t\]*end\[ \t\]+includes?\[ \t\]*\\\*/" \
		$line] {
	    # The above is a liberal match against lines including
	    # the sequence '/* End includes */'
	    break
	}
    }
    close $fileid
    # Sort & remove dependencies
    set worklist [lsort $worklist]
    set ArrInclDep($filename) {} ;# Allow for empty Dep lists
    set ilast {}
    foreach i $worklist {
	if { [string compare $i $ilast] != 0 } {
	    lappend ArrInclDep($filename) [set ilast $i]
	}
    }
    unset ArrInclDep($filename,WoRkInG)
    return $ArrInclDep($filename)
}


######################################################################
# Build utility proc's
proc Prepend { prefix namelist } {
    set newlist ""
    foreach i $namelist { lappend newlist "$prefix$i" }
    return $newlist
}
proc Postpend { namelist suffix } {
    set newlist ""
    foreach i $namelist { lappend newlist "$i$suffix" }
    return $newlist
}
proc Compile { srcfile outfile } {
    global CCC CFLAGS EXTRACFLAGS CHANGECOUNT OUTFLAG RM BROKENCCCERRORS
    # Remove $outfile if it already exists
    if [file exists $outfile] {
       set cmd [concat $RM $outfile]
       puts [join $cmd]
       catch {eval $cmd}
    }

    # Do compile
    set cmd [concat $CCC $CFLAGS $EXTRACFLAGS $srcfile $OUTFLAG$outfile 2>@ stdout]
    set exitcode [catch {BuildExec $cmd} msg]
    # Note: exec returns an error condition if the command writes
    #   to stderr, regardless of the error return from the command
    #   itself.  We (presumably) only want the actual command error
    #   code, so we redirect stderr to stdout.  (In particular, the
    #   Microsoft cl compiler always writes out its header message
    #   to stderr.)
    # Note to Note: Unfortunately, under Windows with Tcl 7.6p2,
    #   the above redirection send stderr directly to the console,
    #   not to msg.  But at least this doesn't raise an error.

    # Now figure out if there was an error
    set errcode 0
    if {!$BROKENCCCERRORS && $exitcode} { set errcode 1 }
    if {![file exists $outfile]} { set errcode 1 }
    if {$errcode} {
	puts "--- Compile messages ---\n$msg\n------------------------"
	puts "Exit code: $exitcode"
    }
    incr CHANGECOUNT
    return $errcode
}
proc Link { objlist outfile } {
    global LINKER CCC LFLAGS LIBXTR CHANGECOUNT RM LINKOUTFLAG
    # Remove $outfile if it already exists
    if {[file exists $outfile]} {
      set cmd [concat $RM $outfile]
      puts [join $cmd]
      catch {eval $cmd}
    }

    # Do link
    set cmd [concat $LINKER $objlist $LFLAGS $LINKOUTFLAG$outfile]
    set errcode 0
    if [catch {BuildExec $cmd} msg] {
	puts "---- Link messages -----\n$msg\n------------------------"
    }
    if {![file exists $outfile] && [string compare $LIBXTR ""]!=0} {
	puts "Retrying link with extra libraries: $LIBXTR"
	if [catch {BuildExec [concat $cmd $LIBXTR]} msg] {
	    puts "---- Link messages -----\n$msg\n------------------------"
	}
    }
    if ![file exists $outfile] {
	flush stdout
	puts stderr "Sorry; Unable to build $outfile"
	set errcode 1
    } else {
	puts "build: \`$outfile\' complete"
    }
    incr CHANGECOUNT
    return $errcode
}
proc DepCheck { targlist deplist } {
# Returns a sublist of targlist that contains all elements of
# targlist that are older than the newest member of deplist
    global PLATFORM
    set deptime -1
    foreach i $deplist {
        # Ought to use file join
        if {![string compare [file dirname $i] .]} {
          if {[string compare $PLATFORM windows]==0} {
            set i [file join [file dirname [info script]] [file tail $i]]
          } else {
            set i [file dirname [info script]]/[file tail $i]
          }
        }
	if [catch {file mtime $i} modtime] {
	    puts "Error: Dependency $i does not exist"
	} else {
	    if ($modtime>$deptime) { set deptime $modtime }
	}
    }
    set buildlist ""
    foreach i $targlist {
	if {[catch {file mtime $i} targtime] || $targtime<$deptime} { 
	    lappend buildlist $i
	}
    }
    return $buildlist
}
proc MakeDirectory {dir} {
  global MKDIR
  if ![file isdirectory $dir] {
    if [file exists $dir] {
	puts stderr "ERROR: Non-directory $dir already exists"
	exit 1
    }
    if [catch {eval $MKDIR $dir} msg] {
	puts stderr "ERROR: Unable to make directory $dir: $msg"
	exit 1
    }
  }
}
proc InstallFiles { filelist directory } {
  # need to check that "ShellPrompt copy" under DOS will take multiple files
  global CP
  if [regexp "\[^ \t\n\]" $filelist] {
    set cmd [concat $CP $filelist $directory]
    puts [join $cmd]
    catch {eval $cmd}
    flush stderr
  }
}
proc InstallDirectoryTree { srcdir destdir } {
  global CP
  puts stderr "\nInstalling $srcdir as $destdir..."
  if {![file exists $srcdir]} {return}
  if {![file isdirectory $srcdir]} {
    puts stderr "ERROR: InstallDirectoryTree needs directories as arguments."
    exit 1
  }
  MakeDirectory $destdir
  set files {}
  set subdirs {}
  # Be sure hidden files and directories get copied too
  foreach item [glob $srcdir/* $srcdir/.*] {
    if {[string match $srcdir/. $item]} {
        continue
    }
    if {[string match $srcdir/.. $item]} {
        continue
    }
    if [file isdirectory $item] {
      if [string compare [file tail $item] CVS] { #Skip CVS directories
        lappend subdirs $item
      }
    }
    if [file isfile $item] {
      lappend files $item
    }
  }
  InstallFiles $files $destdir
  foreach dir $subdirs {
    InstallDirectoryTree $dir $destdir/[file tail $dir]
  }
}

proc MakeSourceDistribution { thisdir destdir } {
  global CP
  puts stderr "\nDistributing $thisdir as $destdir..."
  if {![file exists $thisdir]} {return}
  if {![file isdirectory $thisdir]} {
    puts stderr "ERROR: MakeSourceDistribution needs directories as arguments."
    exit 1
  }
  cd $thisdir
  set savedir [pwd]
  set distribscript distrib.tcl
  if {![file exists $distribscript]} {return}
  if {![file isfile $distribscript]} {return}
  if {![file readable $distribscript]} {return}
  set distributionlist {}
  source $distribscript
#  set distributionlist [Prepend $thisdir/ $distributionlist]
  MakeDirectory $destdir
  set files {}
  set subdirs {}
  foreach item $distributionlist {
    if [file isdirectory $item] {
      lappend subdirs $item
      continue
    }
    if [file isfile $item] {
      lappend files $item
      continue
    }
    puts stderr "Ignoring request to distribute $item ."
    puts stderr "No such file or directory."
  }
  puts stderr "Copying $files to $destdir"
  InstallFiles $files $destdir
  foreach dir $subdirs {
    MakeSourceDistribution $dir $destdir/[file tail $dir]
    cd $savedir
  }
  puts stderr "\nDone distributing $thisdir as $destdir..."
}


# Push the values of a list of variable names onto a stack
proc PushVarList {varlist} {
  global build_stack
  Default build_stack {}
  foreach var $varlist {
    global $var
    lappend build_stack [set $var]
  }
  return [llength $build_stack]
}

# Restore the values retrieved from a stack to a list of variables
proc PopVarList {varlist} {
  global build_stack
  set stacktop [llength $build_stack]
  set stackidx [expr $stacktop - [llength $varlist]]
  set stackpiece [lrange $build_stack $stackidx $stacktop]
  set build_stack [lreplace $build_stack $stackidx $stacktop]
  set i 0
  foreach var $varlist {
    global $var
    set $var [lindex $stackpiece $i]
    incr i
  }
  return [llength $build_stack]
}

proc DoBuild {} {
  global stack argv

  # A list of the variables whose values should not be changed by
  # builds in parent or child directories
  set LOCALVARS {PRESUBDIRS POSTSUBDIRS TARGETS SOURCES EXTINCL BINDIR
                 INCLUDEDIRS OTHEROBJECTS CONFIGLITTER
		 CONFIGSCRIPT CONFIGSOURCE}
  foreach var $LOCALVARS {
    global $var
  }



  global CHANGECOUNT MV RM OBJEXT CCCEXT EXEEXT PLATFORM
  global INSTALLROOT INSTALLEXECROOT INSTALLBIN INSTALLTCLLIB INSTALLDEMOS
  global INCLUDEDIRECTORYFLAG SOURCEDISTRIBUTIONDIRECTORY
  global CFLAGS EXTRACFLAGS DEBUGFLAG STRICTWARNFLAG MACRODEFINEFLAG

  set EXTRACFLAGS {}

  # Do builds in the PRE- subdirectories
  PushVarList $LOCALVARS
  foreach dir $PRESUBDIRS {
    # Ought to use file join here
    EvaluateScript $dir/build.tcl
  }
  PopVarList $LOCALVARS

  set CHANGECOUNT 0
  
  # Build object module list from SOURCES
  set OBJECTS $SOURCES
  
  # Add proper suffix to object files
  set OBJECTS [Postpend $OBJECTS $OBJEXT]
  set TARGETOBJ [Postpend $TARGETS $OBJEXT]
  set OTHEROBJECTS [Postpend $OTHEROBJECTS $OBJEXT]
  
  # Add executable extension (if any)
  set TARGETEXE [Postpend $TARGETS $EXEEXT]
  
  # Be sure files are called by pathnames relative to cwd so building can be
  # performed from any working directory.
  set thisdir [file dirname [info script]]
  foreach list $LOCALVARS {
    if {[string compare $list CONFIGSCRIPT]} {
      set $list [Prepend $thisdir/ [set $list]]
    }
  }

puts "Building '$argv' in $thisdir"
puts "	with Tcl [info patchlevel]"
  
  # Prepend $OBJECTS and $TARGETS with $BINDIR
  set OBJECTS [Prepend $BINDIR/ $OBJECTS]
  set TARGETOBJ [Prepend $BINDIR/ $TARGETOBJ]
  set TARGETEXE [Prepend $BINDIR/ $TARGETEXE]
  
  # Create BINDIR if necessary
  # Only need it if there are things to build
  if [expr [llength $OBJECTS] + [llength $TARGETS]] {
    MakeDirectory $BINDIR
  }
  
  switch -- $argv {
    distclean {
      set files [concat $OBJECTS $TARGETOBJ $TARGETEXE $CONFIGLITTER]
      if [string compare $files {}] {
        set cmd [concat $RM $files]
        puts [join $cmd]
        catch {eval $cmd}
      }
    }
    clean {
      set files [concat $OBJECTS $TARGETOBJ $TARGETEXE]
      if [string compare $files {}] {
        set cmd [concat $RM $files]
        puts [join $cmd]
        catch {eval $cmd}
      }
    }
    configure {
      foreach srcfile $CONFIGSOURCE {
        if {![file exists $srcfile]} {
          puts "build.tcl configure: Can't complete configuration"
          puts "	Missing required file: $srcfile"
          exit 1
        } 
      }
      if {[catch {eval $CONFIGSCRIPT} msg]} {
        puts "build.tcl configure: Configuration failure."
        puts "	$msg"
        exit 1
      }
    }
    install {
      puts stderr "Installing..."
      MakeDirectory $INSTALLROOT
      MakeDirectory $INSTALLEXECROOT
      MakeDirectory $INSTALLBIN
      MakeDirectory $INSTALLTCLLIB
      MakeDirectory $INSTALLDEMOS
      InstallFiles $TARGETEXE $INSTALLBIN
      InstallDirectoryTree \
          [file join [file dirname [info script]] library] $INSTALLTCLLIB
      InstallDirectoryTree \
          [file join [file dirname [info script]] demos] $INSTALLDEMOS
    }
    srcdist {
      if [info exists SOURCEDISTRIBUTIONDIRECTORY] {
        set thisdir [file dirname [info script]]
        MakeSourceDistribution $thisdir $SOURCEDISTRIBUTIONDIRECTORY
      } else {
        puts stderr "The variable SOURCEDISTRIBUTIONDIRECTORY must be defined"
        puts stderr "before you can build a source distribution."
        puts stderr "Edit [file dirname [info script]]/build/local.tcl"
        flush stderr
        exit 1
      }
      # No recursion!  It's built into MakeSourceDistribution.
      return
    }
    debug -
    {} {
      # Check for "debug" request
      if {[string compare $argv debug]==0} {
          set DEBUGLVL 100
          AppendUnique CFLAGS $DEBUGFLAG
          AppendUnique CFLAGS $STRICTWARNFLAG
      } else {
          set DEBUGLVL 0
      }

      # Import the debug level macro definition into the compiler
      AppendUnique CFLAGS ${MACRODEFINEFLAG}DEBUGLVL=$DEBUGLVL

      # Change C++ src to have proper extension for compiler/linker
      foreach srcfile [concat [Postpend $SOURCES .cc] [Postpend $TARGETS .cc]] {
        if {[file exists $srcfile] && [string compare $CCCEXT .cc]} {
          regsub {.cc$} $srcfile $CCCEXT destfile
          set cmd [concat $MV $srcfile $destfile]
          puts [join $cmd]
          catch {eval $cmd}
        }
      }
  
      # Add proper suffix to SOURCES
      set SOURCES [Postpend $SOURCES $CCCEXT]
      set TARGETSRC [Postpend $TARGETS $CCCEXT]

      foreach dir $INCLUDEDIRS {
        AppendUnique EXTRACFLAGS $INCLUDEDIRECTORYFLAG$dir
      }

      # Compile objects
      set CompilationErrors 0
      foreach i [concat $SOURCES $TARGETSRC] {
        set deplist [concat $i [FindIncludeDep $i] $EXTINCL]
        set objfile "$BINDIR/[file tail [file rootname $i]]$OBJEXT"
        if {[DepCheck $objfile $deplist]!=""} {
          incr CompilationErrors [Compile $i $objfile]
        }
      }
  
      # Change C++ src back to have distribution extension
      foreach srcfile [concat $SOURCES $TARGETSRC] {
        if {[file exists $srcfile] && [string compare $CCCEXT .cc]} {
          regsub "$CCCEXT$" $srcfile .cc destfile
          set cmd [concat $MV $srcfile $destfile]
          puts [join $cmd]
          catch {eval $cmd}
        }
      }
  
      if {$CompilationErrors!=0} {
        puts "build: $CompilationErrors Compilation Failure(s)"
        exit 1
      }
  
      # Link objects
      set LinkErrors 0
      foreach trg $TARGETEXE {
        set trgobj [file rootname $trg]$OBJEXT
        set linklist [DepCheck $trg [concat $OBJECTS $trgobj $OTHEROBJECTS]]
        if {"$linklist"!=""} {
          incr LinkErrors [Link [concat $OBJECTS $trgobj $OTHEROBJECTS] $trg]
        }
      }      
      if {$LinkErrors!=0} {
        puts "build: $LinkErrors Link Errors"
        exit 2
      }
  
      if {!$CHANGECOUNT && [string compare $TARGETEXE {}]} {
       puts "build: \`$TARGETEXE\' is up to date."
      }
    }
    default {
      puts "ERROR: unrecognized argument: $argv"
      exit 1
    }
  }
  
  # Do builds in the POST- subdirectories
  PushVarList $LOCALVARS
  foreach dir $POSTSUBDIRS {
    # Ought to use file join here
    EvaluateScript $dir/build.tcl
  }
  PopVarList $LOCALVARS
}

# Temporary placement until better system is in place.
proc BuildOommfPort { outfile } {
    if [file exists $outfile]  {
	puts stdout \
	"BuildOommfPort: Overwriting existing machine header file $outfile"
#	"BuildOommfPort: Using existing machine header file $outfile"
#	return 2
    }

    # Globals
    global CCC CCCEXT LINKOUTFLAG EXEEXT MV tcl_platform
    global RANDOM_GENERATOR_HEADER_BLOCK BINDIR INSTALLBIN

    # See if we can tell what platform we are on
    set systemtype unknown  ;# For local use
    if {![info exists tcl_platform(platform)]} {
	set systemtype unix  ;# Can't tell, so assume unix
    } else {
	if {[string compare $tcl_platform(platform) "unix"] == 0} {
	    set systemtype unix
	} elseif {[string compare $tcl_platform(platform) "windows"] == 0} {
	    set systemtype windows
	} elseif {[string compare $tcl_platform(platform) "macintosh"] == 0} {
	    set systemtype macintosh
	}
    }

    # Local variables, initialize to defaults, but see next block
    if {[string compare $systemtype windows] == 0 } {
	set cc "cl"                    ;# Microsoft compiler
	set linkoutflag "/link /OUT:"  ;# No trailing space allowed
	set cccext      ".cpp"
	set exeext      ".exe"
	set objext      ".obj"
    } else { ;# Assume unix
	set cc "cc"
	set linkoutflag "-o "          ;# Trailing space required
	set cccext      ".cc"
	set exeext      ""
	set objext      ".o"
    }

    # Reset defaults from globals, if the latter exist
    if [info exists CCC] { set cc $CCC }
    if [info exists LINKOUTFLAG] { set linkoutflag $LINKOUTFLAG }
    if [info exists CCCEXT] { set cccext $CCCEXT }
    if [info exists EXEEXT] { set exeext $EXEEXT }
    if [info exists OBJEXT] { set objext $OBJEXT }

    # Prepend the directory of the current build.tcl file to the file names.
    set thisdir [file dirname [info script]]
    set varinfobase "varinfo"
    set varinfoexe [file join $thisdir $varinfobase$exeext]
    set varinfoobj [file join $thisdir $varinfobase$objext]

    # Change C++ src to have proper extension for compiler/linker
    set srcfile [file join $thisdir $varinfobase.cc]
    if {[file exists $srcfile] && [string compare $CCCEXT .cc]} {
      regsub {.cc$} $srcfile $CCCEXT destfile
      set cmd [concat $MV $srcfile $destfile]
      puts [join $cmd]
      catch {eval $cmd}
    }

    set varinfosrc [file join $thisdir $varinfobase$cccext]

    # Build varinfo executable if it doesn't already exist
#    if {![file executable $varinfoexe]} {
      catch {Compile $varinfosrc $varinfoobj}
      catch {Link $varinfoobj $varinfoexe}
#    }

    # Change C++ src back to have distribution extension
    if {[file exists $varinfosrc] && [string compare $CCCEXT .cc]} {
      regsub "$CCCEXT$" $varinfosrc .cc destfile
      set cmd [concat $MV $varinfosrc $destfile]
      puts [join $cmd]
      catch {eval $cmd}
    }

    # Check compilation
    if {![file executable $varinfoexe]} {
      puts stderr "Unable to compile $varinfosrc"
      return 0
    }

    # Run varinfo and parse output
    set varlist [BuildExec $varinfoexe]
    foreach vartype {char short int long float double} {
	set varorder($vartype) -1   ;# Safety
	regexp "$vartype is *(\[0-9\]*) bytes wide *Byte order: *(\[0-9\]*)" \
		$varlist tempmatch varsize($vartype) varorder($vartype)
    }
    set varsize(pointer) -1
    regexp {void \* is *([0-9]*)} $varlist tempmatch varsize(pointer)

    # Open output file
    if { [string compare $outfile stdout] == 0 } {
	set fileid stdout
    } else {
	if [catch {open $outfile w} fileid] {
	    puts stderr \
		    "Unable to open machine header file $outfile for writing"
	    return 0
	}
    }

    # Dump header info
    puts $fileid \
{/* FILE: oommfport.h               -*-Mode: c++-*-
 *
 * Machine specific #define's and typedef's, generated by oommfport.tcl
 *
 */

#ifndef OOMMFPORT_H
#define OOMMFPORT_H

/* End includes */     /* Optional directive to build.tcl */}

    # Include files & defines that may be system dependent
    puts $fileid {
/* Include and defines files that *may* be system dependent. */
#include <stdlib.h>
#include <string.h>
#include <tcl.h>
#define MEMBLOCKSIZE 4096 /* Natural system memory blocksize.  Shouldn't  */
               /* affect program correctness, but may affect performance. */
}
    # Check that the header file tcl version agrees with
    # the running version of tclsh
    if {![regexp -- {^([0-9]+).([0-9]+)([abpi]?)([0-9]*)$} [info patchlevel] \
	    junk major minor level serial]} {
	puts stderr \
		"ERROR: Bad patchlevel info obtained in oommfport.tcl"
	return 0
    }
    if { [string compare $level "a"]==0 } {
	set level 0
    } elseif { [string compare $level "b"]==0 } {
	set level 1
    } else {
	set level 2
    }
    if { [string compare $serial {}]==0 } {
	set serial 0
    }
    puts $fileid "
/* Define the Tcl release level and serial if missing from tcl.h */
#ifndef TCL_RELEASE_LEVEL
#define TCL_RELEASE_LEVEL $level
#endif
#ifndef TCL_RELEASE_SERIAL
#define TCL_RELEASE_SERIAL $serial
#endif
#ifndef TCL_PATCH_LEVEL
#define TCL_PATCH_LEVEL	   \"[info patchlevel]\"
#endif

/* Check that the version of Tcl used to produce this file    */
/* agrees with the version number from tcl.h at compile time. */
/* Remove this check at your own risk!!!                      */
#if (TCL_MAJOR_VERSION  != $major) || \\
    (TCL_MINOR_VERSION  != $minor) || \\
    (TCL_RELEASE_LEVEL  != $level) || \\
    (TCL_RELEASE_SERIAL != $serial)
#error tclsh ($major.$minor.$level.$serial) and tcl.h versions differ
#endif
"
    set shell omfsh$EXEEXT
    puts $fileid "
/* The filename of the Tcl shell of the OOMMF project.         */
/* It is used to interpret Tcl scripts in sub-processess.      */
#define OMF_SHELL \"$shell\"
"
    set ssp $INSTALLBIN
    lappend ssp [file join [pwd] [string trim $BINDIR ./]]
    puts $fileid "
/* The portion of the search path for the OOMMF project's Tcl  */
/* shell which is determined at configure/build time.          */
#define OMF_SHELL_STATIC_SEARCH_PATH \"$ssp\"
"

    # Write universal (I hope!) typedef's
    puts $fileid "/* Variable type declarations.  The '****#m' */"
    puts $fileid "/* types are at *least* '#' bytes wide.      */"
    puts $fileid "typedef  int                 BOOL;"
    puts $fileid "typedef  unsigned char       BYTE;"
    puts $fileid "typedef  char                CHAR;"
    puts $fileid "typedef  signed char        SCHAR;"
    puts $fileid "typedef  unsigned char      UCHAR;"

    # Write float typedef's
    if { $varsize(float) != $varsize(double) } {
	puts $fileid "typedef  float              REAL$varsize(float);"
    }
    puts $fileid "typedef  double             REAL$varsize(double);"
    if { $varsize(float) >= 4 } {
	puts $fileid "typedef  float              REAL4m;"
    }
    if { $varsize(float) >= 8 } {
	puts $fileid "typedef  float              REAL8m;"
    } elseif { $varsize(double) >= 8 } {
	puts $fileid "typedef  double             REAL8m;"
    }
    # May add "long double" support in the future

    # Write integer typedef's
    if { $varsize(short) < $varsize(int) } {
	puts $fileid "typedef  short               INT$varsize(short);"
	puts $fileid "typedef  unsigned short     UINT$varsize(short);"
    }
    puts $fileid "typedef  int                 INT$varsize(int);"
    puts $fileid "typedef  unsigned int       UINT$varsize(int);"
    if { $varsize(long) > $varsize(int) } {
	puts $fileid "typedef  long                INT$varsize(long);"
	puts $fileid "typedef  unsigned long      UINT$varsize(long);"
    }
    foreach msize { 2 4 8 16 } {
	if { $varsize(int) >= $msize } {
	    # Use type "int" if possible, as this is likely
	    # to be the preferred machine word size
	    puts $fileid "typedef  int                 INT${msize}m;"
	    puts $fileid "typedef  unsigned int       UINT${msize}m;"
	} elseif { $varsize(long) >=$msize } {
	    # Otherwise, fall back on long type
	    puts $fileid "typedef  long                INT${msize}m;"
	    puts $fileid "typedef  unsigned long      UINT${msize}m;"
	}
	# In the future, may add support for "long long" integral type
    }

    # Pointers
    puts $fileid "\n#define POINTERWIDTH $varsize(pointer)"

    # Byte order.  For now just use 4-byte wide ordering
    foreach vartype { int long short float double } {
	if { $varsize($vartype) == 4 } {
	    puts $fileid "#define BYTEORDER $varorder($vartype)"
	    break
	}
    }

    # Machine platform types
    puts $fileid "\n/* System type info */"
    puts $fileid "#define UNIX 1"
    puts $fileid "#define WINDOWS 2"
    puts $fileid "#define MACINTOSH 3"
    # Note: Local variable "systemtype" is set at top of this proc
    if {[string compare $systemtype "unix"] == 0} {
	puts $fileid "#define SYSTEM_TYPE UNIX"
    } elseif {[string compare $systemtype "windows"] == 0} {
	puts $fileid "#define SYSTEM_TYPE WINDOWS"
    } elseif {[string compare $systemtype "macintosh"] == 0} {
	puts $fileid "#define SYSTEM_TYPE MACINTOSH"
    }

    # Random number generator prototypes and macros.
    puts $fileid $RANDOM_GENERATOR_HEADER_BLOCK

    # Window vs. unix -isms
    if {[string compare $systemtype unix] == 0} {
	puts $fileid {
/* For unix */
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>  /* Child process cleanup */

/* Directory path separator; Unix uses '/', DOS uses '\'. */
/* Rumor has it that the Mac uses ':'.                    */
#define DIRSEPCHAR '/'
#define DIRSEPSTR  "/"
#define PATHSPLITSTR ":"
}
    } elseif {[string compare $systemtype windows] == 0} {
	puts $fileid {
/* For Windows */
#define rint(x) ((x)>=0 ? double(int(x+0.5)) : double(int(x-0.5)))

/* A few definitions taken from Windows' include files */
/* Certainly it's overkill */

#ifdef __cplusplus
extern "C" {
#endif

/* from windef.h */
#ifndef CONST
#define CONST const
#endif

/* from winnt.h */
#if (defined(_M_MRX000) || defined(_M_IX86) || defined(_M_ALPHA) || defined(_M_PPC)) && !defined(MIDL_PASS)
#define DECLSPEC_IMPORT __declspec(dllimport)
#else
#define DECLSPEC_IMPORT
#endif

typedef void *PVOID;
typedef wchar_t WCHAR;
typedef WCHAR *LPCWSTR;
typedef CHAR *LPSTR;
typedef CONST CHAR *LPCSTR;

#ifdef STRICT
#define DECLARE_HANDLE(name)  struct name##__ { int unused; };  typedef struct name##__ *name
#else  /* STRICT */
typedef PVOID HANDLE;
#define DECLARE_HANDLE(name)  typedef HANDLE name
#endif /* !STRICT */

/* from windef.h */
#undef far
#undef near

#define far
#define near
#if (_MSC_VER >= 800) || defined(_STDCALL_SUPPORTED)
#define WINAPI __stdcall
#define APIENTRY WINAPI
#else
#define WINAPI
#define APIENTRY WINAPI
#endif

#define FAR far
#define NEAR near

typedef unsigned int UINT;

#ifndef WIN_INTERNAL
DECLARE_HANDLE (HWND);
#endif
DECLARE_HANDLE(HINSTANCE);

/* from winuser.h */
#if !defined(_USER32_)
#define WINUSERAPI DECLSPEC_IMPORT
#else
#define WINUSERAPI
#endif

/* Message Box Support (partial) */
#define MB_OK                       0x00000000L
#define MB_ICONHAND                 0x00000010L
#define MB_ICONEXCLAMATION          0x00000030L
#define MB_ICONSTOP                 MB_ICONHAND
#define MB_TASKMODAL                0x00002000L
#define MB_SETFOREGROUND            0x00010000L

#ifndef UNICODE_ONLY
WINUSERAPI
int
WINAPI
MessageBoxA(
    HWND hWnd ,
    LPCSTR lpText,
    LPCSTR lpCaption,
    UINT uType);
#endif //!UNICODE_ONLY
#ifndef ANSI_ONLY
WINUSERAPI
int
WINAPI
MessageBoxW(
    HWND hWnd ,
    LPCWSTR lpText,
    LPCWSTR lpCaption,
    UINT uType);
#endif //!ANSI_ONLY
#ifdef UNICODE
#define MessageBox  MessageBoxW
#else
#define MessageBox  MessageBoxA
#endif // !UNICODE
WINUSERAPI
BOOL
WINAPI MessageBeep(UINT uType);
void WINAPI ExitProcess(UINT);

#ifdef __cplusplus
} // end of 'extern "C"'
#endif

/* Directory path separator; Unix uses '/', DOS uses '\'. */
/* Rumor has it that the Mac uses ':'.                    */
#define DIRSEPCHAR '\\'
#define DIRSEPSTR  "\\"
#define PATHSPLITSTR ";"
}
    }

    # Dump trailer
    puts $fileid "#endif /* OOMMFPORT_H */"

    if { [string compare $outfile stdout] != 0 } {
	close $fileid
    }
}

