/* FILE: ptsearch.h                 -*-Mode: c++-*-
 *
 * Classes for searching unordered lists of points in space.
 */

#ifndef PTSEARCH_H
#define PTSEARCH_H

#endif // PTSEARCH_H

#define BOX_REFINELEVEL 5    // Default box refinement level.
/// 4**BOX_REFINELEVEL >= # of search boxes, and probably closer to
/// = than to > :-)   NOTE: 4**5 = 1024

#include "machine.h"
#include "template.h"
#include "vector.h"

/* End includes */     // This is an optional directive to build.tcl,
                       // that excludes the remainder of the file from
                       // '#include ".*"' dependency building

//////////////////////////////////////////////////////////////////////////
// Base LocatedVector element.
struct BoxElt {
  INT4m select_count;      // Convenience variable, incremented by
  /// BoxList::GetClosest2D(...), used to keep track of number of
  /// times this element has been included in a client list.
  LocatedVector lv;
  BoxElt() { select_count=0; }
  BoxElt(const LocatedVector &_lv) { select_count=0; lv=_lv; }

  // A couple convenience functions
  REAL4 DistSq2D(const BoxElt &other) const {
    REAL4 xdelta=lv.location.x-other.lv.location.x;
    REAL4 ydelta=lv.location.y-other.lv.location.y;
    return xdelta*xdelta+ydelta*ydelta;
  }
  REAL4 DistSq2D(const Vec3f &v) const {
    REAL4 xdelta=lv.location.x-v.x;
    REAL4 ydelta=lv.location.y-v.y;
    return xdelta*xdelta+ydelta*ydelta;
  }
};


//////////////////////////////////////////////////////////////////////////
// Atomic LocatedVector* list element.
class BoxList;  // Forward declaration
class Box {
private:
  static const ClassDoc class_doc;
  REAL4 xmin,ymin,xmax,ymax;  // Box extents
  List<BoxElt *> *list;       // Comparison list
  void FillList(List<BoxElt *> *biglist); // Fills "list" with a refined
  /// version of biglist using both Sup & L2 norms; also sets incount.
  INT4m incount;              // Number of elements of list inside box
  BOOL IsIn2D(REAL4 x,REAL4 y) const;
public:
  Box() { list=NULL; xmin=ymin=xmax=ymax=0.0; incount=0; }
  ~Box() { if(list!=NULL) delete list; }
  void InitNoRefine(REAL4 _xmin,REAL4 _ymin,REAL4 _xmax,REAL4 _ymax,
		    List<BoxElt> *wholelist);
  /// Used for initialization of the first box *only*.  Sets up pointers
  /// in List<BoxElt *> list to the entire List<BoxElt> biglist WITHOUT
  /// refinement.  Also sets incount.

  void Init(REAL4 _xmin,REAL4 _ymin,REAL4 _xmax,REAL4 _ymax,
	    Box &bigbox);
  /// Called by BoxList::Refine(INT4m), this routine first deletes
  /// List<BoxElt *> list, if it exists, and then copies those elements of
  /// List<BoxElt *> biglist that could be "closest 2D position" to current
  /// box, as defined by above box extents.  Also sets incount.

  BOOL IsIn2D(const Vec3f &pos) const;  /// True if pos is inside box

  REAL4m GetBoxArea() const;   // Returns (xmax-xmin)*(ymax-ymin)
  INT4m GetSpaceWaste() const; // Returns size of unused list space (bytes)

  int InCount(void) const { return incount; }
  int ListCount(void) const { return list->GetSize(); }
  BoxElt* GetClosest2D(const Vec3f &pos);
};

//////////////////////////////////////////////////////////////////////////
// Client interface to a list of BoxElt, Box* box, which is a list of
// lists of LocatedVector*'s.  Also holds List<BoxElt> wholelist, which is
// the list of LocatedVector's pointed into by entries of Box* box.
class BoxList
{
private:
  static const ClassDoc class_doc;
  REAL4 xmin,ymin,xmax,ymax; // Whole box region
  INT4m Nx,Ny;               // Subbox counts, x & y
  REAL4 xdelta,ydelta;       // Box cell dimensions
  Box* box;                  // Array, size=Nx*Ny
  BOOL box_valid;            // True if current box refinement is valid.
  List<BoxElt> *wholelist;
  /// It is into this list that the lists of box[] point.
  REAL4 aveincount,avelistcount; // Box averages; set by ::Refine()
  void MakeBoxValid();       // Convenience routine.  Does nothing
  /// if box_valid is true.  Otherwise, calls ::Refine(), and if
  /// DEBUGLVL is large enough then issues a warning message.
  void RefineOnce();         // Called as needed from ::Refine()
  void InitBox();            // Should only be called when box==NULL
public:
  void ClearList();  // Releases *box, empties wholelist, and zeros
  /// box region x/ymin, x/ymax.

  BoxList();
  ~BoxList();

  INT4m GetSize() const { return wholelist->GetSize(); }
  REAL4 GetAveInCount() { MakeBoxValid(); return aveincount; }
  REAL4 GetAveListCount() { MakeBoxValid(); return avelistcount; }
  INT4m GetBoxCount() const { return Nx*Ny; }
  INT4m GetSpaceWaste() const;

  void AddPoint(const LocatedVector &lv);
  /// BoxList forms an internal list (wholelist) of each
  ///  LocatedVector passed in, since the last call to
  ///  ::ClearList().  Pointers returned ::GetClosest point
  ///  into this list.  Also the List<BoxElt *> lists in each
  ///  box[] point into this same list.
  /// NOTE: AddPoint automatically enlarges the box region as
  ///  necessary to include the point lv.  The client should
  ///  call SetBoxRegion after all calls to AddPoint if it
  ///  desires different behavior.

  void GetBoxRegion(REAL4 &_xmin,REAL4 &_ymin,REAL4 &_xmax,REAL4 &_ymax) const
  {
    _xmin=xmin; _xmax=xmax; _ymin=ymin; _ymax=ymax;
  }
  void AddMarginBoxRegion(REAL4 xmargin,REAL4 ymargin) {
    xmin-=xmargin; xmax+=xmargin; ymin-=ymargin; ymax+=ymargin;
    box_valid=0;
  }
  void InflateBoxRegion(REAL4 xscale,REAL4 yscale) {
    REAL4 xoldsize=xmax-xmin;
    REAL4 xnewsize=xoldsize*xscale;
    REAL4 yoldsize=ymax-ymin;
    REAL4 ynewsize=yoldsize*yscale;
    AddMarginBoxRegion((xnewsize-xoldsize)/2.,(ynewsize-yoldsize)/2.);
  }
  void SetBoxRegion(REAL4 _xmin,REAL4 _ymin,REAL4 _xmax,REAL4 _ymax) {
    xmin=_xmin; xmax=_xmax; ymin=_ymin; ymax=_ymax;
    box_valid=0;
  }
  void ExpandBoxRegion(REAL4 _xmin,REAL4 _ymin,REAL4 _xmax,REAL4 _ymax) {
    if(_xmin<xmin) xmin=_xmin;
    if(_xmax>xmax) xmax=_xmax;
    if(_ymin<ymin) ymin=_ymin;
    if(_ymax>ymax) ymax=_ymax;
    box_valid=0;
  }

  void DeleteRefinement();
  /// Pitches any existing *box array, and marks box_valid=1.

  INT4m Refine(INT4m refinelevel=BOX_REFINELEVEL);
  ///  If box_valid=1, then refines existing *box structure refinelevel
  /// times.  This will usually increase the number of boxes in *box by
  /// a factor of 4**refinelevel.
  ///  If box_valid=0, then ::Refine first throws away any existing *box
  /// array, and issues a warning message (for DEBUGLVL sufficiently
  /// large) to aid program efficiency development.
  ///  In all cases, box_valid is set to 1 upon exit.
  /// The return value is the total number of boxes Nx*Ny, which with
  /// luck will fit inside an INT4m.  Overflow of INT4m will likely
  /// occur at refinelevel of about 16, at which point you are probably
  /// out of memory; if not, everything will break down as the indexing
  /// into box[] fails.

  INT4m Refine(INT4m boxcount,REAL4m ave_incount,REAL4m ave_listcount);
  /// An adaptive interface to the RefineOnce routine.  Does successive
  /// refinements until one of the following conditions is met:
  ///        number of boxes (=Nx*Ny) >= import boxcount 
  ///   or        average box incount <  import ave_incount
  ///   or      average box listcount <  import ave_listcount
  /// Note that each refinement step increases the number of boxes
  /// by a factor or 2 or 4, so the number of boxes may be up to
  /// 4 times larger than the import boxcount on return.  This can
  /// be reduced to a factor of 2 by judicious choice of boxcount.
  /// Example: If boxcount<=1024, then your are guaranteed of getting
  /// at most 2048 boxes, but if boxcount=1025, then you may get as
  /// many as 4096.
  ///   You may turn off any of these checks by setting the
  /// corresponding import to 0.  Setting all 3 to 0 raises an error.
  ///   The return value is Nx*Ny.

  INT4m ClearSelectCounts();
  INT4m GetClosest2D(const Vec3f &pos,
		     LocatedVector* &lv,INT4m &select_count);
  /// Given position pos, returns in lv the element of the wholelist
  /// that is closest to pos, ignoring the z component.  The value of
  /// selected_count is the number of times this element has been
  /// selected (via calls to GetClosest2D) since the last call to
  /// ::ClearSelectCounts (which is also effectively called at BoxList
  /// instantiation time), including this one.  In particular, each
  /// call to this function will increment the select_count field of
  /// one BoxElt, and the parameter select_count will always return >=1.
  ///  NOTE: If GetClosest2D is called at a time when box_valid is
  ///        false, then ::Refine() will be called, using the default
  ///        BOX_REFINELEVEL value.  A warning message will be raised
  ///        in this case if DEBUGLVL is large enough.  Client routines
  ///        are encouraged to explicitly call ::Refine() as necessary.
  ///        The ::Refine() call is time consuming, so it is best if the
  ///        programmer is fully aware of when it is called.

  void GetWholeList(List<LocatedVector> &list);
  /// Fills export "list" with pointers to each entry of ::wholelist.
  /// Any existing elements of list are cleared.
  /// Increments select_count field of each entry of ::wholelist by one.
};

// USUAL BOXLIST USAGE:
//     // Initialize boxlist
//     BoxList boxlist;
//     LOOP1:
//       < Generate LocatedVector lv >
//       boxlist.AddPoint(lv);
//       < More points?  Yes -> goto LOOP1 >
//     boxlist.InflateBoxRegion(1.2,1.2);   // Add a 10% margin on all sides
//     Refine();
//     ...
//     // Use boxlist to generate a sub-sampled list of points
//     List<LocatedVector *> sublist;  LocatedVector *lvp;
//     boxlist.ClearSelectCounts();  // Not required for first access
//     LOOP2:
//       < Generate sample point Vec3f pos >
//       boxlist.GetClosest2D(pos,lvp,count);
//       if(count==1) sublist.Append(lvp);
//       < More sample points?  Yes -> goto LOOP2 >
//     < Use sublist as desired >
//     ...
//     < When sublist & boxlist are _both_ not needed, the space
//      in boxlist can be freed via boxlist.ClearList().         >
//

