/* FILE: vector.cc                -*-Mode: c++-*-
 *
 * Vector and collections of vectors
 *
 */

#include <stdio.h>

#include "vector.h"

/* End includes */     // This is an optional directive to build.tcl,
                       // that excludes the remainder of the file from
                       // '#include ".*"' dependency building

//////////////////////////////////////////////////////////////////////////
// 4-byte floating point vectors
const ClassDoc Vec3f::class_doc("Vec3f",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","11-Dec-1996");

//////////////////////////////////////////////////////////////////////////
// 3D rectangular solid
const ClassDoc BoundingBox::class_doc("BoundingBox",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","13-Dec-1996");

BOOL BoundingBox::IsIn(const Vec3f &location) const
{
#define MEMBERNAME "IsIn"
  return (location.Bigger(min) && (max.Bigger(location)));
#undef MEMBERNAME
}

void BoundingBox::SortAndSet(const Vec3f &corner1,const Vec3f &corner2) {
  if(corner1.x<corner2.x) { min.x=corner1.x; max.x=corner2.x; }
  else                    { min.x=corner2.x; max.x=corner1.x; }
  if(corner1.y<corner2.y) { min.y=corner1.y; max.y=corner2.y; }
  else                    { min.y=corner2.y; max.y=corner1.y; }
  if(corner1.z<corner2.z) { min.z=corner1.z; max.z=corner2.z; }
  else                    { min.z=corner2.z; max.z=corner1.z; }
}

void BoundingBox::Set(const Vec3f &corner1,const Vec3f &corner2) {
  min.x=corner1.x; max.x=corner2.x;
  min.y=corner1.y; max.y=corner2.y;
  min.z=corner1.z; max.z=corner2.z;
}

BOOL BoundingBox::ExpandWith(const Vec3f &include_point)
{ // Expands box as necessary to include 'include_point'. Returns
  // 1 if expansion occurred, 0 otherwise.
  if(IsEmpty()) {
    min=max=include_point;
    return 1;
  }
  BOOL expand=0;
  if(include_point.x<min.x) { min.x=include_point.x; expand=1; }
  if(include_point.y<min.y) { min.y=include_point.y; expand=1; }
  if(include_point.z<min.z) { min.z=include_point.z; expand=1; }
  if(include_point.x>max.x) { max.x=include_point.x; expand=1; }
  if(include_point.y>max.y) { max.y=include_point.y; expand=1; }
  if(include_point.z>max.z) { max.z=include_point.z; expand=1; }
  return expand;
}

BOOL BoundingBox::ExpandWith(const BoundingBox& other_box)
{ // Expands box as necessary to include 'other_box'.  Returns 1 if
  // expansion occurred, 0 otherwise
  if(IsEmpty()) {
    if(other_box.IsEmpty()) return 0;
    Set(other_box);
    return 1;
  }
  BOOL expand=0;
  if(other_box.min.x<min.x) { min.x=other_box.min.x; expand=1; }
  if(other_box.min.y<min.y) { min.y=other_box.min.y; expand=1; }
  if(other_box.min.z<min.z) { min.z=other_box.min.z; expand=1; }
  if(other_box.max.x>max.x) { max.x=other_box.max.x; expand=1; }
  if(other_box.max.y>max.y) { max.y=other_box.max.y; expand=1; }
  if(other_box.max.z>max.z) { max.z=other_box.max.z; expand=1; }
  return expand;
}

BoundingBox IntersectBoundingBoxes(const BoundingBox& box1,
				   const BoundingBox& box2)
{ // Returns the box resulting from the intersection of box1 and box2.
  // If the two imports have empty intersection, then the returned
  // box has .min=(0,0,0), .max=(-1,-1,-1)
  BoundingBox box;
  box.Reset();  // Safety
  if(box1.IsEmpty() || box2.IsEmpty()) return box;
  Vec3f min,max;
  min.x=MAX(box1.min.x,box2.min.x);
  min.y=MAX(box1.min.y,box2.min.y);
  min.z=MAX(box1.min.z,box2.min.z);
  max.x=MIN(box1.max.x,box2.max.x);
  max.y=MIN(box1.max.y,box2.max.y);
  max.z=MIN(box1.max.z,box2.max.z);
  box.Set(min,max);
  if(box.IsEmpty()) box.Reset();
  return box;
}

Vec3f BoundingBox::GetDiag(void) const { 
  if(IsEmpty()) return Vec3f(0.,0.,0.);
  return max-min;
}
