/* FILE: mifio.h           -*-Mode: c++-*-
 *
 * Header file for I/O routines involving standardized
 * Micromagnetic Input Form (MIF) data.
 */

#ifndef MIFIO_H
#define MIFIO_H

#define MIFVECDIM 3

#include "zeeman.h"

/* End includes */

class FieldRangeArray {
private:
  int RangeCount;         // Total number of ranges
  double* StartField;     // Start value for each range
  double* StopField;      // Stop value for each range
  int* StepCount;         // Number of steps in each range
  int RangesFilled;       // Number of ranges currently setup
  int StepTotal;          // Total number of steps
  int CurIndex;           // Working field index; 0<=CurIndex<=StepTotal,
                          //  NB: Number of indices = Number of steps + 1
  void Free(void);
  int AreSameField(double* field1,double* field2); // Returns 1 if
                                                   // field1==field2.
public:
  FieldRangeArray(void);
  void FreeSpace(void) { Free(); }
  void MakeSpace(int rangecount);
  int AddRange(const double* start,const double* stop,int steps);
  int AddRange(const double* onefield=NULL);
  int AddRange(const char** param); // Should be 2*MIFVECDIM+1 strings.
  int ConsistencyCheck(void);
  int SetFieldIndex(int newindex);  // Note: newindex<0 => newindex=last index
  void StepIndex(int& errorcode);
  /// errorcode=0 on success, -1 if already at last step
  int GetStepTotal() { return StepTotal; }
  int GetCurrentIndex() { return CurIndex; }
  double* CurrentField(void);
  double* Field(int i);
  ~FieldRangeArray(void);
};

enum MifAnisotropyType { mif_uniaxial, mif_cubic, mif_anis_notset };
enum MifGeometryType { mif_rectangle, mif_ellipse, mif_oval, mif_mask, mif_geom_notset };
enum MifDemagnetizationType { mif_slowpipe, mif_fastpipe, mif_2dslab,
			      mif_3dslab, mif_3dcharge,mif_nodemag,
mif_demag_notset };

// Micromagnetic Input Form
class MIF {
private:
  String InName; // Name of input file
  // Material parameters
  double Ms;    // Saturation magnetization in A/m;
  double A;     // Exchange stiffness in J/m;
  double K1;    // First anisotropy constant, in J/m^3;
  MifAnisotropyType AnisType;
  double AnisDir1[MIFVECDIM];  // Primary anisotropy direction;
  double AnisDir2[MIFVECDIM];  // Secondary anisotropy direction;
                               // for some AnisTypes.
  char** AnisInitArgv;  // Mag init routine name and parameters.
  int    AnisInitArgc;  // Number of parameters, including name

  // Demag specifications
  MifDemagnetizationType DemagType;
  // Part geometry
  double PartWidth;   // Nominal part width in meters.
  double PartHeight;  // Nominal part height in meters.
  double PartThickness; // Part thickness in meters.  Required for
                        // some DemagTypes.
  double CellSize;    // Cell size in meters.  May affect PartWidth and
                      // PartHeight.
  MifGeometryType GeomType;
  String GeomTypeParam;  // Shape specific parameter
  // Initial magnetization specification
  char** MagInitArgv;  // Mag init routine name and parameters.
  int    MagInitArgc; // Number of parameters, including name

  // Experiment parameters
  FieldRangeArray FieldRange;
  Zeeman* AppliedField;

  // Output Specifications
  String OutBaseName;
  int LogLevel;

  // Miscellaneous
  int RandomizerSeed;
  double InitialIncrement;     // in _radians_
  int UserInteractionLevel;
  int UserReportCode;
  int MaxIterationCount;

  // Helper functions
  void Init(void);
  void CleanUp(void);
  void SetDefaults(void);
  int  ReadLine(FILE* infile,int max_strc,
		  int &strc,String* strv);
  void AdjustAnisDirs(void);
  int  ConsistencyCheck(void);

public:
  MIF(void);
  ~MIF(void);
  void Read(char* inname,int &errorcode);
  const char* GetOutBaseName(void) { return OutBaseName.GetStr(); }
  int GetLogLevel(void) { return LogLevel; }
  const char* GetDemagName(void);
  double GetPartWidth(void)     { return PartWidth; }
  double GetPartHeight(void)    { return PartHeight; }
  double GetPartThickness(void) { return PartThickness; }
  double GetCellSize(void)      { return CellSize; }
  MifAnisotropyType GetAnisType(void) { return AnisType; }
  const double* GetAnisDir(int i);
  const char* GetAnisInitName(void) { return AnisInitArgv[0]; }
  const char* const* GetAnisInitArgv(void) {
    return (const char* const*)AnisInitArgv;
  }
  int GetAnisInitArgc(void) { return AnisInitArgc; }
  MifGeometryType GetGeometry(String &param)
  { param=GeomTypeParam; return GeomType; }
  char** GetMagInitArgv(void) { return MagInitArgv; }
  double GetMs(void) { return Ms; }
  double GetA(void)  { return A; }
  double GetK1(void) { return K1; }

  Zeeman* GetExternalFieldObject(void) { return AppliedField; }
  void TransferExternalField(Zeeman*  &MyAppliedField) {
    // Releases ownership
    MyAppliedField=AppliedField; AppliedField=NULL;
  }
  int GetFieldStepIndex() { return FieldRange.GetCurrentIndex(); }
  int GetFieldStepTotal() { return FieldRange.GetStepTotal(); }
  /// NOTE: This is the total number of *steps*.  The total range
  ///  of indices is 0<=index<=GetFieldStepTotal(), i.e.,
  ///  GetFieldStepTotal()+1 field states.
  void ResetField(void) {
    // Resets FieldRange and AppliedField
    FieldRange.SetFieldIndex(0);
    if(AppliedField!=NULL)
      AppliedField->SetNomFieldVec3D(Ms,Vec3D(GetNomExternalField()));
  }
  int StepField(int &errorcode) {
    // Steps FieldRange and updates AppliedField
    FieldRange.StepIndex(errorcode);
    if(errorcode==0 && AppliedField!=NULL)
      AppliedField->SetNomFieldVec3D(Ms,Vec3D(GetNomExternalField()));
    return GetFieldStepIndex();
  }
  int SetFieldStepIndex(int newindex) {
    FieldRange.SetFieldIndex(newindex);
    if(AppliedField!=NULL)
      AppliedField->SetNomFieldVec3D(Ms,Vec3D(GetNomExternalField()));
    return GetFieldStepIndex();
  }
  double* GetNomExternalField(void) { return FieldRange.CurrentField(); }
  double* GetNomExternalField(int i) { return FieldRange.Field(i); }

  double GetInitialIncrement(void) { return InitialIncrement; }
                    // Maximum initial increment in _radians_.
  int GetRandomizerSeed(void) { return RandomizerSeed; }
  int GetUserInteractionLevel(void) { return UserInteractionLevel; }
  int GetUserReportCode(void) { return UserReportCode; }
  int GetMaxIterationCount(void) { return MaxIterationCount; }
  void DumpPartialInfo(FILE* fptr);
  int IsValid() { return !ConsistencyCheck(); }
};

#endif // .NOT. MIFIO_H

