/* FILE: vecio.h             -*-Mode: c++-*-
 *
 * This file contains class declarations vector input/output.
 *
 */

#ifndef VECIO_H
#define VECIO_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* End includes */

#define VFH_BADMAGIC      1
#define VFH_UNKNOWNENDIAN 2
#define VFH_BADOFFSET     3
#define VFH_BADNOTE       4
#define VFH_BADVALUES     5
#define VF_PREMATUREEOF   6

extern const char VecFileMagic02[8];
extern const double VecFileEndianTest;

// Class Polygon defines a closed (not necessarily convex) polygon.
// It is used by the bitmap drawing routines.
struct TwoPoint {
public:
  double x,y;
  double Mag(void) { return hypot(x,y); }
};

class Polygon {
private:
  int count;
  TwoPoint *vertex;
  int CrossSegment(TwoPoint& Q,TwoPoint& P1,TwoPoint& P2);
public:
  Polygon(void);
  Polygon(int vertex_count);
  Polygon(Polygon& poly); // Implicit copy constructor
  ~Polygon(void);
  void Resize(int new_vertex_count);
  TwoPoint& operator[](int index);
  Polygon& operator=(Polygon& poly);
  int IsInside(TwoPoint& pt);  // Returns 1 if pt is inside the polygon
  void Scale(double xscale,double yscale);    // Coordinate
  void Shift(double xoffset,double yoffset);  //    transformations
  void Rotate(double angcos,double angsin);   //  "  "
  void AdjustFirstToInteger(void); // Rotates and rescales bitmap
  // as necessary to shift vertex[0] to the nearest integral coordinate
};

struct FileHeader {
 public:
          FileHeader(void);
  BYTE    Magic[8];
  REAL8   EndianTest;
  INT4    Nx,Ny,Nz,dimen,dataoffset;
  BYTE    Reserved[36];
  BYTE    Note[448];
  int     Read(FILE* fptr,int& swapinbytes);
  int     Write(FILE* fptr);
  int     Verify(void);
};

class FileVector {
 private:
  int          SwapInBytes; // 1 => swap byte order on binary input
  FILE*        fptr;
  FileHeader   fh;
  int          Nx,Ny,Nz,dimen,dataoffset;
// Gnu gcc wants these definitions to be inside the class declaration,
// but some other C++ compilers only accept them outside...
// See also the file vecio.cc
#ifdef __GNUC__
  const int    ArrowVertexCount=8;
  const double ArrowLineWidth=0.05;
  const double ArrowLength=0.90;
  const double ArrowHeadRatio=0.375;
  const double VertMagCrossover=0.1;
  const int    CrossVertexCount=12;
  const int    MaxColorValue=255;
  const double BitmapCircleRadius=0.25;
  const double BitmapDotRadius=0.05;
  const double BitmapLineWidth=0.08;  
  const double BitmapArrowHeadRatio=0.45;
  const double BitmapArrowMinPixels=3.;
  const double PageWidth=(8.5*72);
  const double PageHeight=(11.*72);
  const double PageLeftMargin=(1.0*72);
  const double PageTopMargin= (1.0*72);
  const int    MinGrey=0;
  const int    MaxGrey=255;
#else // !__GNUC__
  static const int    ArrowVertexCount;
  static const double ArrowLineWidth;
  static const double ArrowLength;
  static const double ArrowHeadRatio;
  static const double VertMagCrossover; // If inplane magnetization is
               // smaller than this, draw as an arrow tail or point.
  static const int    CrossVertexCount;
  static const int    MaxColorValue;  // Maximum color value
  static const double BitmapCircleRadius;
  static const double BitmapDotRadius;
  static const double BitmapLineWidth;  
  static const double BitmapArrowHeadRatio;
  static const double BitmapArrowMinPixels;
  // PostScript output constants
  static const double PageWidth;  // Paper page width in points
  static const double PageHeight; // Paper page height in points
  static const double PageLeftMargin; // Left and right margin in points
  static const double PageTopMargin;  // Top and bottom margin in points
  // Greylevel output constants
  static const int MinGrey;
  static const int MaxGrey;
#endif // __GNUC__
  // Bitmap output variables and routines
  int          cellsize; // Cell size for bitmapped output (in pixels)
  int          bmwidth,bmheight;      // Bitmap dimensions
  int**        bitmap;  // Bitmap
  Polygon basearrow,workarrow,basecross;
  void DrawBitmapLine(int r1,int c1,int r2,int c2,int color);
  void DrawBitmapArrow(int xcoor,double len,
		       double angcos,double angsin,int color);
  void DrawBitmapCircleCross(int xcoor,int color);
  void DrawBitmapCircleDot(int xcoor,int color);
  void DrawBitmapDot(int xcoor,int color);

  // PostScript output variables
  int          TitleSize; // Title font size in points

  // GreyGif output variables
  double minval,maxval;

 public:
  FileVector(int psfontsize=19); // psfontsize=PostScript font size in points
  ~FileVector(void);
  void    SetNote(BYTE *newnote);
  const BYTE* GetNote(void) { return fh.Note; }
  int     ReadHeader(const char *filename,
		     int &Xdimen,int &Ydimen,int &Zdimen,int &Vecdimen);
  int     ReadVecs(int count,double* arr);
  int     WriteHeader(const char *filename,
		      int Xdimen,int Ydimen,int Zdimen,int Vecdimen);
  int     WriteVecs(int count,double* arr);
  int     WritePSHeader(const char *filename,
			double Xdimen,double Ydimen,double Zdimen,
			int Vecdimen,
			int drawtitle=1,char *titleappendix=NULL);
  int     WritePSVecs(int count,double* arr,int comp0,int comp1,
		      int compcolor);
  int     WritePSTrailer(void);
  int     WritePPMHeader(const char *filename,
			 int Xdimen,int Ydimen,int Zdimen,
			 int Vecdimen,int Imagecellsize);
  int     WritePPMVecs(int count,double* arr,int comp0,int comp1,
		      int compcolor,int colormap=0);
  int     WriteColorPPMHeader(const char *filename,
			 int Xdimen,int Ydimen,int Zdimen,
			 int Vecdimen,double max);
  int     WriteColorPPMPix(int count,double* arr);
  void    WriteColorPPMClose(void); // Call this _INSTEAD_ of Close()
  void    Close(void);
};

#endif /* .NOT. VECIO_H */
