# FILE: counter.tcl
#
# An example of the usage of the em_class command to create a new
# Tcl command 'omf_counter' which exhibits some object-oriented
# features, including private vs. public access control to procedures
# and variables.

# Find a way to get this information into the class:
#
# NAME:		omf_counter
# SYNOPSIS:	package require OmfCounter 0.0
# SYNOPSIS:	omf_counter ?attributes?
# DESCRIPTION:	Each invocation of omf_counter creates a new instance of a
# DESCRIPTION:	counter object and returns the name of a new command which
# DESCRIPTION:	may be used to query, modify, and destroy that instance.  
# DESCRIPTION:	Additional configuration attributes may be specified to the 
# DESCRIPTION:	omf_counter or via the config method of the instance 
# DESCRIPTION:	command.

# AUTHOR, MAINTAINER, EMAIL, GLOBAL VARS USED...
# author/maintainer/contact
# date
# organization
# usage
# description
# web site
# global litter
# copyright info
# license info
# other documentation, possibly categorized by audience
#
# debugging enable/disable/level

em_class omf_counter {

  public variable 	type = integer {
    if {($type != "integer") && ($type != "roman") && ($type != "alphabet") \
        && ($type != "list")} {
      error "legal values: alphabet integer list roman"
    }
  }

  public variable 	initial	= 1 {
    if {![regexp {^[0-9]+$} $initial] || ($initial == 0)} {
      error "value must be a positive integer"
    }
  }

  public variable 	increment = 1 {
    if {![regexp {^[0-9]+$} $increment] || ($increment == 0)} {
      error "value must be a positive integer"
    }
  }

  public variable 	case = upper {
    if {($case != "upper") && ($case != "lower")} {
      error "legal values: upper lower"
    }
  }

  public variable	list = {}

  variable value

  constructor {args} {
    eval $this configure $args
    $this reset
  }

  method reset {} {value initial} {
    set value $initial
    return [$this get]
  }

  method next {} {value increment} {
    set retval [$this get]
    incr value $increment
    return $retval
  }

  private method get {} {type value case list} {
    switch -- $type {
      alphabet {
        switch -- $case {
          upper {
            return [string toupper [$class integerToAlphabet $value]]
          }
          lower {
            return [$class integerToAlphabet $value]
          }
        }
      }
      integer {
        return $value
      }
      list {
        return [lindex $list [expr {$value % [llength $list]}]]
      }
      roman {
        switch -- $case {
          upper	{
            return [string toupper [$class integerToRoman $value]]
          }
          lower {
            return [$class integerToRoman $value]
          }
        }
      }
    }
  }

  # Convert an integer to its Roman numeral representation.
  proc integerToRoman {avalue} {} {
    if {![regexp {^[0-9]+$} $avalue] || ($avalue == 0)} {
      error "
'$class integerToRoman' error:
  	Invalid value: $avalue

usage:
	$class integerToRoman <value>

Invalid call:
	$class integerToRoman $avalue

<value> must be a positive integer"
    }
    if {$avalue > 3999} {
      error "
'$class integerToRoman' error:
  	Invalid value: $avalue

usage:
	$class integerToRoman <value>

Invalid call:
	$class integerToRoman $avalue

<value> exceeds range translatable to Roman numerals"
    }
    set result ""
    # Handle any thousands
    if {[string length $avalue] == 4} {
      for {set i 0} {$i < [string index $avalue 0]} {incr i} {
        append result m
      }
      set avalue [expr $avalue - $i * 1000]
    }
    set elements "mdclxvi"
    # Handle the rest
    for {set digits 3} {$digits > 0} {incr digits -1} {
      if {$digits == [string length $avalue]} {
        append result [$class romanDigitMap [string index $avalue 0] \
            [string range $elements [expr 6 - 2 * $digits] \
            [expr 8 - 2 * $digits] ] ]
        regsub {^[0-9]} $avalue {} avalue
      }
    }
    return $result
  }

  # Convert a positive number value to an alphabetic representation.
  # Use a-z as "digits" of base 26 system, sort of. (There is no zero).
  # Assumes a-z contiguous in character set, like ASCII
  proc integerToAlphabet {avalue} {} {
    if {![regexp {^[0-9]+$} $avalue] || ($avalue == 0)} {
      error "
'$class integerToAlphabet' error:
  	Invalid value: $avalue

usage:
	$class integerToAlphabet <value>

Invalid call:
	$class integerToAlphabet $avalue

<value> must be a positive integer"
    }
    set result {}
    incr avalue -1
    if {$avalue > 25} {
      set l [expr $avalue / 26]
      set result [$class integerToAlphabet $l]
      set avalue [expr $avalue % 26]
    }
    append result [format %c [expr $avalue + \
		[set v 0; scan a %c v; set v]]]
    return $result
  }

  # Convert a decimal digit to a Roman numeral representation,
  # using the symbols in the 3-character string symbols as
  # the 10, 5, and 1 representer.
  private proc romanDigitMap {digit symbols} {} {
    set i [string index $symbols 2]
    set v [string index $symbols 1]
    set x [string index $symbols 0]
    switch $digit {
      0	{return ""}
      1	{return $i}
      2	{return $i$i}
      3	{return $i$i$i}
      4	{return $i$v}
      5	{return $v}
      6	{return $v$i}
      7	{return $v$i$i}
      8	{return $v$i$i$i}
      9	{return $i$x}
    }
  }
}
