/* FILE: omfsh.cc                 -*-Mode: c++-*-
 *
 *	The Tcl shell for the OOMMF project.
 *
 */

#include <tcl.h>
#include <tk.h> 
#include <string.h>

#include "misc.h"

/* End includes */     

static int RegisterCommand(Tcl_Interp* interp,
			   const char* name,Tcl_CmdProc* cmd)
{
  // I don't know that it is necessary to copy name to a
  // writeable buffer, but better safe than sorry.
  char buf[512];
  strncpy(buf,name,sizeof(buf));  buf[sizeof(buf)-1]='\0';
  if(Tcl_CreateCommand(interp,buf,cmd,NULL,NULL)==NULL) {
    fprintf(stderr,"Unable to register command ->%s<- "
	    "with Tcl Interpreter\n",buf);
    fflush(stderr);
    exit(1);
  }
  return TCL_OK;
}


static int
Omfsh_Init(Tcl_Interp *interp)
{
  // Register miscellaneous C/C++ routines with interpreter
  RegisterCommand(interp,"CullZombies",CullZombies);
  return TCL_OK;
}

int Tcl_AppInit(Tcl_Interp *interp)
{
  if (Tcl_Init(interp)   == TCL_ERROR ||
      Omfsh_Init(interp) == TCL_ERROR ||
      Oommf_Init(interp) == TCL_ERROR   )
    {
      return TCL_ERROR;
    }
  return TCL_OK;
}

int Tk_AppInit(Tcl_Interp *interp)
{
  if (Tcl_Init(interp)   == TCL_ERROR ||
      Tk_Init(interp)    == TCL_ERROR ||
      Omfsh_Init(interp) == TCL_ERROR ||
      Oommf_Init(interp) == TCL_ERROR   )
    {
      return TCL_ERROR;
    }

  // Setup default icon on root window
  const char *cmd="after idle {OMF_SetIcon .}";
  char *buf=new char[strlen(cmd)+1];
  strcpy(buf,cmd);
  Tcl_Eval(interp,buf);
  delete[] buf;
  
  return TCL_OK;
}

static void remove_arg(int &argc,char **argv,int remove_index)
{ // Removes argument "remove_index" from argument list by
  // translating all arguments after it down one spot and
  // decrementing argc.
  if(remove_index<0 || remove_index>=argc) return;
  for(int j=remove_index+1;j<argc;j++) {
    argv[j-1]=argv[j];
  }
  argc--;
}

static void ExtractOptions(int &argc,char **argv,
			   BOOL &forkflag,BOOL &tkflag)
{
  forkflag=0; // Default is to not fork a child process (if possible)
  tkflag=1;   // Default is to use tk    

  int i=1;
  while(i<argc) {
    if(strcmp(argv[i],"--")==0){
      // End flag processing
      remove_arg(argc,argv,i);
      break;
    }

    if(argv[i][0]!='-') {
      // Not an option flag. Under Windows, should one
      // use '/' instead of '-' ??? 
      i++;
      continue;
    }

    char *optflag=&(argv[i][1]);
    if(*optflag=='-') optflag++;  // Treat "-" and "--" the same

    // Known options
    if(strcmp(optflag,"tk")==0) {
      tkflag=1;   remove_arg(argc,argv,i);
    }
    else if(strcmp(optflag,"notk")==0 || strcmp(optflag,"no_tk")==0) {
      tkflag=0;   remove_arg(argc,argv,i);
    }
    else if(strcmp(optflag,"fork")==0) {
      forkflag=1; remove_arg(argc,argv,i);
    }
    else if(strcmp(optflag,"nofork")==0 || strcmp(optflag,"no_fork")==0) {
      forkflag=0; remove_arg(argc,argv,i);
    }
    else {
      // Anything else is left on the command line untouched, for
      // possible use by the shell script
      i++;
    }
  }
}

#if (SYSTEM_TYPE==WINDOWS)
static void ParseParams(char *paramstr,char *exename,
			int argv_size,int &argc,char** &argv)
{ // The command line does not get parsed (tokenized) under Windows
  // (unlike unix) so this routine provides such to ease portability
  // between Windows and unix.  One NOTE: Under windows, the command
  // name is _NOT_ included in the parameter list!
  // Imports:  argv is a buffer allocated in the client of length
  //   (at least) argv_size.  paramstr is the string holding the
  //   untokenized list.
  // Exports: argc is the number of tokens, which will be <=argv_size,
  //   and argv[] will hold pointers into paramstr to those tokens.
  //   If argc<argv_size, then argv[argc] is set to NULL.  Also,
  //   argv[0] is set to exename.
  // Note: This routine modifies paramstr[] (through strtok()).
  const char* whitespace=" \t\n\r\f";
  char *cptr=strtok(paramstr,whitespace);
  argc=0;
  argv[argc++]=exename;
  while(argc<argv_size) {
    argv[argc]=cptr;
    if(cptr==NULL) break;
    argc++;
    cptr=strtok(NULL,whitespace);
  }
}

int APIENTRY WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,
		     LPSTR lpszCmdLine,int nCmdShow)
{
  char exename[64];  // Needs to be long enough to hold exename,
  /// but must also be writable to keep and Tcl_Main, Tk_Main happy.
  /// I'd rather not use heap space (via "new") because I don't
  /// know if this gets automatically cleaned up under Windows
  /// (Tcl_Main and Tk_Main calls typically don't return.)
  strncpy(exename,"omfsh",sizeof(exename));
  exename[sizeof(exename)-1]='\0';  // Safety

  int argc;
  char *argv[64];  // Shouldn't need anywhere near this large
  char **stupid=&(argv[0]); // MS C++ compiler (version 4)
  /// apparently doesn't want to convert from a char *[32]
  /// to a char **& in a parameter list.  But this is okay...
  ParseParams(lpszCmdLine,exename,sizeof(argv),argc,stupid);

  BOOL forkflag,tkflag;
  ExtractOptions(argc,argv,forkflag,tkflag);
  // forkflag ignored under Windows

  if(tkflag)
    Tk_Main(argc, argv, Tk_AppInit);
  else
    Tcl_Main(argc, argv, Tcl_AppInit);

  // NOTE: I don't know if Windows automatically releases "new"
  //   (heap) memory on program exit.  Since Tk_Main doesn't return,
  //   I can't "delete" any heap memory after this point, so I'll
  //   just hope that argv & exename above get cleaned up by the system.

  return 0;
}
#elif (SYSTEM_TYPE == UNIX)
int main(int argc,char **argv)
{
  BOOL forkflag,tkflag;
  ExtractOptions(argc,argv,forkflag,tkflag);

#if DEBUGLVL<1
  if(!forkflag || fork()==0) { // Fork only occurs if forkflag!=0
    // Child process
    freopen("/dev/null","r",stdin);
    freopen("/dev/null","r",stdout);
#endif // DEBUGLVL<1

    if(tkflag)
      Tk_Main(argc, argv, Tk_AppInit);
    else
      Tcl_Main(argc, argv, Tcl_AppInit);
    return 0;

#if DEBUGLVL<1
  }
#endif // DEBUGLVL<1

  return 0;
}
#else
#error "Unknown SYSTEM_TYPE"
#endif // SYSTEM_TYPE == WINDOWS
